from typing import Any, Dict, Optional, Union

from pymongo.collection import Collection

product_collection: Collection = None
coupon_collection: Collection = None
discount_collection: Collection = None
user_collection: Collection = None
user_profile_collection: Collection = None
private_profile_collection: Collection = None
cart_collection: Collection = None
site_collection: Collection = None
checkout_collection: Collection = None
order_collection: Collection = None
counter_collection: Collection = None
state_collection: Collection = None


def set_collections(
    product: Collection,
    coupon: Collection,
    discount: Collection,
    user: Collection,
    user_profile: Collection,
    private_profile: Collection,
    cart: Collection,
    site: Collection,
    checkout: Collection,
    order: Optional[Collection] = None,
    counter: Optional[Collection] = None,
    state: Optional[Collection] = None,
) -> None:
    global product_collection
    global coupon_collection
    global discount_collection
    global user_collection
    global user_profile_collection
    global private_profile_collection
    global cart_collection
    global site_collection
    global checkout_collection
    global order_collection
    global counter_collection
    global state_collection
    product_collection = product
    coupon_collection = coupon
    discount_collection = discount
    user_collection = user
    user_profile_collection = user_profile
    private_profile_collection = private_profile
    cart_collection = cart
    site_collection = site
    checkout_collection = checkout
    order_collection = order
    counter_collection = counter
    state_collection = state


def get_collection(name: str) -> Collection:
    if name == "product":
        return product_collection
    if name == "coupon":
        return coupon_collection
    if name == "discount":
        return discount_collection
    if name == "user":
        return user_collection
    if name == "cart":
        return cart_collection
    if name == "user_profile":
        return user_profile_collection
    if name == "private_profile":
        return private_profile_collection
    if name == "site":
        return site_collection
    if name == "checkout":
        return checkout_collection
    if name == "order":
        return order_collection
    if name == "counter":
        return counter_collection
    if name == "state":
        return state_collection


def get_site_config_doc() -> Union[Dict[str, Any], None]:
    if site_collection is None:
        return None
    return site_collection.find_one({"name": "config"})  # type: ignore


def get_site_config(key: str, default: Any = None) -> Any:
    config_doc = get_site_config_doc()
    if config_doc is None:
        return default
    return config_doc.get(key, default)



from fastapi import Request


async def debug_logger(request: Request) -> None:
    """
    Log the HTTP request to the console.

    According to this discussion: https://github.com/tiangolo/fastapi/issues/394,
    consuming request data in middleware is problematic, so we shouldn't put this kind of
    logger into the middleware.

    Instead, put in the route's dependency list like this:
    `app.include_router(api_router, dependencies=[Depends(log_json)])`
    """
    msg = ["\n*************"]
    msg += ("[URL]", str(request.url))
    try:
        data = await request.json()
        if data:
            msg += ("[JSON]", str(data))
        form_data = await request.form()
        if form_data:
            msg += ("[FORM DATA]", str(form_data))
    except Exception:
        pass
    msg += ("*************\n",)
    print("\n".join(msg))



# ecpay
from typing import Any, Dict, List

from fastel.cart.datastructures import InvoiceTypes, PaymentSubTypes
from fastel.payment.ecpay.models.invoice import IssueB2BModel, IssueB2CModel

ECPAY_CALLBACK_SUBTYPE = {
    "Credit": PaymentSubTypes.credit,
    "ATM": PaymentSubTypes.atm,
    "WebATM": PaymentSubTypes.webatm,
    "CVS": PaymentSubTypes.cvs,
    "BARCODE": PaymentSubTypes.barcode,
}

CARRIER_TYPE_TABLE = {
    InvoiceTypes.B2C: "",
    InvoiceTypes.B2C_DONATE: "",
    InvoiceTypes.B2C_PROVIDER: "1",
    InvoiceTypes.B2C_NPC: "2",
    InvoiceTypes.B2C_PHONE_CARRIER: "3",
}


def _replace_limit_name(name: str, limit: int) -> str:
    if len(name) > limit:
        return name[: limit - 3] + "..."
    return name


def generate_B2B_invoice_data(order: Dict[str, Any]) -> Dict[str, Any]:
    items = order.get("items", [])
    extra_items = order.get("extra_items", [])
    discount_items = order.get("discount_items", [])
    invoice_data = {
        "RelateNumber": order.get("order_number", ""),
        "CustomerIdentifier": order.get("b2b_company_no", ""),
        "CustomerAddress": order.get("buyer_city", "")
        + order.get("buyer_district", "")
        + order.get("buyer_address"),
        "CustomerEmail": order.get("buyer_email", ""),
        "TaxType": "1",
        "SalesAmount": order.get("sales", 0),
        "TaxAmount": order.get("tax", 0),
        "TotalAmount": order.get("total", 0),
        "Items": generate_B2B_invoice_item(items, extra_items, discount_items),
        "InvType": "07",
    }
    return IssueB2BModel.validate(invoice_data).dict(exclude_none=True)


def generate_B2B_invoice_item(
    items: List[Dict[str, Any]],
    extra_items: List[Dict[str, Any]],
    discount_items: List[Dict[str, Any]],
) -> List[Dict[str, Any]]:
    result = []
    seq = 1
    for item in items:
        item_result = {
            "ItemSeq": seq,
            "ItemName": _replace_limit_name(name=item.get("name", ""), limit=256),
            "ItemCount": item["config"].get("qty", 1),
            "ItemWord": item["config"]["extra_data"]
            and item["config"]["extra_data"].get("word", "份")
            or "份",
            "ItemPrice": item.get("unit_sales", 0),
            "ItemAmount": item.get("sales_amount", 0),
        }
        result.append(item_result)
        seq += 1
    # 計算運費
    for extra_item in extra_items:
        item_result = {
            "ItemSeq": seq,
            "ItemName": _replace_limit_name(name=extra_item.get("name", ""), limit=256),
            "ItemCount": 1,
            "ItemWord": "份",
            "ItemPrice": round(extra_item.get("amount", 0) / 1.05, 7),
            "ItemAmount": round(extra_item.get("amount", 0) / 1.05, 7),
        }
        result.append(item_result)
        seq += 1
    # 計算折扣金額
    for discount_item in discount_items:
        item_result = {
            "ItemSeq": seq,
            "ItemName": _replace_limit_name(
                name=discount_item.get("name", ""), limit=256
            ),
            "ItemCount": 1,
            "ItemWord": "份",
            "ItemPrice": -discount_item.get("sale_amount", 0),
            "ItemAmount": -discount_item.get("sale_amount", 0),
        }
        result.append(item_result)
        seq += 1
    return result


def generate_B2C_invoice_data(order: Dict[str, Any]) -> Dict[str, Any]:
    items = order.get("items", [])
    extra_items = order.get("extra_items", [])
    discount_items = order.get("discount_items", [])
    carrier_type = CARRIER_TYPE_TABLE[order["invoice_type"]]
    if order["invoice_type"] == InvoiceTypes.B2C:
        print = "1"
        donation = "0"
        carrier_num = ""
        love_code = None
    elif order["invoice_type"] == InvoiceTypes.B2C_PROVIDER:
        print = "0"
        donation = "0"
        carrier_num = ""
        love_code = None
    elif order["invoice_type"] == InvoiceTypes.B2C_NPC:
        print = "0"
        donation = "0"
        carrier_num = order.get("b2c_npc_code", "")
        love_code = None
    elif order["invoice_type"] == InvoiceTypes.B2C_PHONE_CARRIER:
        print = "0"
        donation = "0"
        carrier_num = order.get("b2c_phone_carrier_code", "")
        love_code = None
    elif order["invoice_type"] == InvoiceTypes.B2C_DONATE:
        print = "0"
        donation = "1"
        love_code = order.get("b2c_donate_code", "")
        carrier_num = ""
    else:
        print = "0"
        donation = "0"
        carrier_num = ""
        love_code = None
    invoice_data = {
        "RelateNumber": order.get("order_number", ""),
        "CustomerName": order.get("buyer_name", ""),
        "CustomerAddr": order.get("buyer_city", "")
        + order.get("buyer_district", "")
        + order.get("buyer_address"),
        "CustomerEmail": order.get("buyer_email", ""),
        "Print": print,
        "Donation": donation,
        "LoveCode": love_code,
        "CarrierType": carrier_type,
        "CarrierNum": carrier_num,
        "TaxType": "1",
        "SalesAmount": order.get("total", 0),
        "Items": generate_B2C_invoice_item(items, extra_items, discount_items),
        "InvType": "07",
    }
    return IssueB2CModel.validate(invoice_data).dict(exclude_none=True)


def generate_B2C_invoice_item(
    items: List[Dict[str, Any]],
    extra_items: List[Dict[str, Any]],
    discount_items: List[Dict[str, Any]],
) -> List[Dict[str, Any]]:
    result = []
    for item in items:
        item_result = {
            "ItemName": _replace_limit_name(name=item.get("name", ""), limit=100),
            "ItemCount": item["config"].get("qty", 1),
            "ItemWord": item["config"]["extra_data"]
            and item["config"]["extra_data"].get("word", "份")
            or "份",
            "ItemPrice": item.get("price", 0),
            "ItemAmount": item.get("amount", 0),
        }
        result.append(item_result)
    # 計算運費
    for extra_item in extra_items:
        item_result = {
            "ItemName": _replace_limit_name(name=extra_item.get("name", ""), limit=100),
            "ItemCount": 1,
            "ItemWord": "份",
            "ItemPrice": extra_item.get("amount", 0),
            "ItemAmount": extra_item.get("amount", 0),
        }
        result.append(item_result)
    # 計算折扣金額
    for discount_item in discount_items:
        item_result = {
            "ItemName": _replace_limit_name(
                name=discount_item.get("name", ""), limit=100
            ),
            "ItemCount": 1,
            "ItemWord": "份",
            "ItemPrice": -discount_item.get("amount", 0),
            "ItemAmount": -discount_item.get("amount", 0),
        }
        result.append(item_result)
    return result



from enum import Enum
from typing import Any, Literal, Optional

from pydantic import BaseModel, validator

LogisticsTypeOptions = Literal["CVS", "Home"]

LogisticsSubTypeOptions = Literal[
    "FAMI",
    "UNIMART",
    "HILIFE",
    "FAMIC2C",
    "UNIMARTC2C",
    "HILIFEC2C",
    "OKMARTC2C",
    "TCAT",
    "ECAN",
]
LogisticsStatusOptions = Literal[
    "pending",
    "in_delivery",
    "delivered",
    "exception",
    "center_delivered",
    "store_delivered",
]

DistanceOptions = Literal["00", "01", "02"]
TemperatureOptions = Literal["0001", "0002", "0003"]
SpecificationOptions = Literal["0001", "0002", "0003", "0004"]
ScheduledPickupTimeOptions = Literal["1", "2", "3", "4"]
ScheduledDeliveryTimeOptions = Literal["1", "2", "3", "4", "12", "13", "23"]

CvsMapSubTypeOptions = Literal[
    "FAMI", "UNIMART", "FAMIC2C", "UNIMARTC2C", "HILIFE", "HILIFEC2C", "OKMARTC2C"
]


class EcpayLogisticsModel(BaseModel):
    MerchantTradeNo: str
    LogisticsType: LogisticsTypeOptions = "Home"
    LogisticsSubType: LogisticsSubTypeOptions = "TCAT"
    GoodsName: Optional[str]
    GoodsAmount: int
    SenderName: str
    SenderPhone: str
    SenderCellPhone: Optional[str]
    ReceiverName: str
    ReceiverCellPhone: str
    ReceiverEmail: str = ""
    IsCollection: Literal["Y", "N"] = "N"
    CollectionAmount: Optional[int]
    ReceiverStoreID: Optional[str]
    TradeDesc: Optional[str]
    Remark: Optional[str]
    ServerReplyURL: str

    # if LogisticsType = Home
    Distance: Optional[DistanceOptions]
    Temperature: Optional[TemperatureOptions]
    Specification: Optional[SpecificationOptions]
    ScheduledPickupTime: Optional[ScheduledPickupTimeOptions]
    ScheduledDeliveryTime: Optional[ScheduledDeliveryTimeOptions]
    ScheduledDeliveryDate: Optional[str]
    PackageCount: Optional[int]
    SenderZipCode: Optional[str]
    SenderAddress: Optional[str]
    ReceiverZipCode: Optional[str]
    ReceiverAddress: Optional[str]

    # only C2C
    ReturnStoreID: Optional[str]
    LogisticsC2CReplyURL: Optional[str]


class LogisticsTypes(str, Enum):
    HOME = "HOME"
    CVS = "CVS"


class EcpayLogisticsCallbackModel(BaseModel):
    MerchantID: str
    MerchantTradeNo: str
    RtnCode: str
    RtnMsg: str
    AllPayLogisticsID: str
    LogisticsType: LogisticsTypes
    LogisticsSubType: str
    GoodsAmount: int
    UpdateStatusDate: str = ""
    ReceiverName: str = ""
    ReceiverPhone: str = ""
    ReceiverCellPhone: str = ""
    ReceiverEmail: str = ""
    ReceiverAddress: str = ""
    CVSPaymentNo: str = ""
    CVSValidationNo: str = ""
    BookingNote: str = ""
    CheckMacValue: str

    @validator("RtnMsg", "ReceiverAddress", "ReceiverName")
    def validate_msg(cls, v: str, **kwargs: Any) -> str:
        return v.encode("Latin-1").decode()


class CVSSubtypes(str, Enum):
    FAMI = "FAMI"
    UNIMART = "UNIMART"
    HILIFE = "HILIFE"
    FAMIC2C = "FAMIC2C"
    UNIMARTC2C = "UNIMARTC2C"
    HILIFEC2C = "HILIFEC2C"
    OKMARTC2C = "OKMARTC2C"


class EcpayCVSMapModel(BaseModel):
    LogisticsType: Literal["CVS"] = "CVS"
    LogisticsSubType: CvsMapSubTypeOptions = "FAMI"
    IsCollection: Literal["Y", "N"]
    Device: int = 0
    ServerReplyURL: str


class CVSMapCallback(BaseModel):
    MerchantID: str = ""
    MerchantTradeNo: str = ""
    LogisticsSubType: str = ""
    CVSStoreID: str = ""
    CVSStoreName: str = ""
    CVSAddress: str = ""
    CVSTelephone: str = ""
    CVSOutSide: str = ""


class EcpayPrintOrderInfo(BaseModel):
    LogisticsSubType: Literal[
        "FAMIC2C",
        "UniMartC2C",
        "HILIFEC2C",
        "OKMARTC2C",
        "FAMI",
        "UNIMART",
        "UNIMARTFREEZE",
        "HILIFE",
        "TCAT",
        "ECAN",
    ]
    AllPayLogisticsID: str
    CVSPaymentNo: Optional[str]
    CVSValidationNo: Optional[str]



import base64
import hashlib
import json
import random
import socket
import string
import struct
import time
from typing import Any, Dict, Optional, Tuple

from Crypto.Cipher import AES

from fastel.exceptions import APIException
from fastel.logistics.sf import ierror


class SHA1:
    """计算消息签名接口"""

    def getSHA1(
        self, token: str, timestamp: str, nonce: str, encrypt: str
    ) -> Tuple[int, Optional[str]]:
        """用SHA1算法生成安全签名
        @param token:  access token
        @param timestamp: 时间戳
        @param encrypt: 密文
        @param nonce: 随机字符串
        @return: 安全签名
        """
        try:
            sortlist = [token, timestamp, nonce, encrypt]
            sortlist.sort()
            sha = hashlib.sha1()
            sha.update("".join(sortlist).encode())
            return ierror.WXBizMsgCrypt_OK, sha.hexdigest()
        except (RuntimeError):
            # print e
            return ierror.WXBizMsgCrypt_ComputeSignature_Error, None

    def getSHA2(
        self, token: str, timestamp: str, nonce: str, encrypt: str
    ) -> Tuple[int, Optional[str]]:
        """用SHA256算法生成安全签名
        @param token:  access token
        @param timestamp: 时间戳
        @param encrypt: 密文
        @param nonce: 随机字符串
        @return: 安全签名
        """
        try:
            sortlist = [token, timestamp, nonce, encrypt]
            sortlist.sort()
            sha = hashlib.sha256()
            # print("shaStr:"+"".join(sortlist))
            sha.update(("".join(sortlist)).encode())
            return ierror.WXBizMsgCrypt_OK, sha.hexdigest()
        except (RuntimeError):
            # print e
            return ierror.WXBizMsgCrypt_ComputeSignature_Error, None


class PKCS7Encoder:
    """提供基于PKCS7算法的加解密接口"""

    block_size = 32

    def encode(self, text: bytes) -> bytes:
        """对需要加密的明文进行填充补位
        @param text: 需要进行填充补位操作的明文
        @return: 补齐明文字符串
        """
        text_length = len(text)
        # 计算需要填充的位数
        amount_to_pad = self.block_size - (text_length % self.block_size)
        if amount_to_pad == 0:
            amount_to_pad = self.block_size
        # 获得补位所用的字符
        pad = chr(amount_to_pad)
        return text + str.encode(pad * amount_to_pad)

    def decode(self, decrypted: bytes) -> bytes:
        """删除解密后明文的补位字符
        @param decrypted: 解密后的明文
        @return: 删除补位字符后的明文
        """
        pad = decrypted[-1]
        # print("pad:"+str(pad))
        if pad < 1 or pad > 32:
            pad = 0
        return decrypted[:-pad]


class Prpcrypt:
    """提供接收和推送给公众平台消息的加解密接口"""

    def __init__(self, key: bytes) -> None:
        # self.key = base64.b64decode(key+"=")
        self.key = key
        # 设置加解密模式为AES的CBC模式
        self.mode = AES.MODE_CBC

    def encrypt(self, text: str, appKey: str) -> Tuple[int, Optional[str]]:
        """对明文进行加密
        @param text: 需要加密的明文
        @return: 加密得到的字符串
        """
        # 16位随机字符串添加到明文开头
        textBytes = (
            str.encode(self.get_random_str())
            + struct.pack("I", socket.htonl(len(text.encode())))
            + str.encode(text)
            + str.encode(appKey)
        )
        # print("需要加密的明文:"+text)
        # 使用自定义的填充方式对明文进行补位填充
        pkcs7 = PKCS7Encoder()
        textBytes = pkcs7.encode(textBytes)
        # print("补位:"+text)
        # 加密
        cryptor = AES.new(self.key, self.mode, self.key[:16])
        try:
            ciphertext = cryptor.encrypt(textBytes)
            # 使用BASE64对加密后的字符串进行编码
            # print("密文:"+bytes.decode(base64.b64encode(ciphertext)))
            return ierror.WXBizMsgCrypt_OK, bytes.decode(base64.b64encode(ciphertext))
        except (RuntimeError):
            # print e
            return ierror.WXBizMsgCrypt_EncryptAES_Error, None

    def decrypt(self, text: str, appKey: str) -> Tuple[int, Optional[str]]:
        """对解密后的明文进行补位删除
        @param text: 密文
        @return: 删除填充补位后的明文
        """
        try:
            cryptor = AES.new(self.key, self.mode, self.key[:16])
            # 使用BASE64对密文进行解码，然后AES-CBC解密
            plain_text = cryptor.decrypt(base64.b64decode(text))
            # print("解密:"+bytes.decode(plain_text))
        except (RuntimeError):
            # print e
            return ierror.WXBizMsgCrypt_DecryptAES_Error, None
        try:
            # 去掉补位字符串
            pkcs7 = PKCS7Encoder()
            content = pkcs7.decode(plain_text)
            # print("随机字符串:"+str(content))
            # print("struct.unpack:"+str(struct.unpack("I",content[16: 20])))
            xml_len = socket.ntohl(struct.unpack("I", content[16:20])[0])
            # print("xml_len:"+str(xml_len))
            xml_content = bytes.decode(content[20 : xml_len + 20])
            # print("xml_content:"+xml_content)
            from_appKey = bytes.decode(content[xml_len + 20 :])
            # print("from_appKey:"+from_appKey)
        except (RuntimeError):
            # print e
            return ierror.WXBizMsgCrypt_IllegalBuffer, None
        if from_appKey != appKey:
            return ierror.WXBizMsgCrypt_ValidateAppKey_Error, None
        # print("明文:"+str(xml_content))
        return 0, xml_content

    def get_random_str(self) -> str:
        """随机生成16位字符串
        @return: 16位字符串
        """
        rule = string.ascii_letters + string.digits
        str = random.sample(rule, 16)
        return "".join(str)


class SFCryptor:
    # 构造函数
    # @param access token: 接口获取的token
    # @param aes_key: 应用上的EncodingAESKey
    # @param app_key: 应用的AppKey
    def __init__(
        self, aes_key: str, app_key: str, access_token: Optional[str] = None
    ) -> None:
        try:
            self.key = base64.b64decode(aes_key + "=")
            assert len(self.key) == 32
        except Exception:
            raise APIException(
                status_code=400,
                error="AES_Encoding_error",
                detail="SF EncodingAESKey unvalid !",
            )
        self.token = access_token
        self.app_key = app_key

    def encrypt_msg(
        self, msg: str, nonce: str, timestamp: Optional[str]
    ) -> Tuple[int, Optional[str], Optional[str]]:
        # 将公众号回复用户的消息加密打包
        # @param msg: 需加密的报文
        # @param timestamp: 时间戳，可以自己生成,如为None则自动用当前时间
        # @param nonce: 随机串，可以自己生成
        # encrypt_msg: 加密后的可以直接回复用户的密文，包括msg_signature, timestamp, nonce, encrypt的xml格式的字符串,
        # return：成功0，encrypt_msg,失败返回对应的错误码None
        if not self.token:
            raise APIException(
                status_code=400, error="SF_Encrypt_Token_not_found", detail=""
            )
        pc = Prpcrypt(self.key)
        ret, encrypt = pc.encrypt(msg, self.app_key)
        if not encrypt or ret != 0:
            return ret, encrypt, None
        if timestamp is None:
            timestamp = str(int(time.time()))
        # 生成安全签名
        sha1 = SHA1()
        ret, signature = sha1.getSHA2(self.token, timestamp, nonce, encrypt)
        if not signature or ret != 0:
            return ret, None, signature
        # print("密文："+encrypt+",签名:"+signature)
        return ret, encrypt, signature

    def decrypt_msg(self, post_data: str) -> Tuple[int, Dict[str, Any]]:
        # 获取解密后的明文
        # @param post_data: 密文
        #  xml_content: 解密后的原文，当return返回0时有效
        # @return: 成功0，失败返回对应的错误码
        # 验证安全签名
        pc = Prpcrypt(self.key)
        ret, decrypt_data = pc.decrypt(post_data, self.app_key)
        if decrypt_data:
            result = json.loads(decrypt_data)
            assert isinstance(result, dict)
            return ret, result
        else:
            return ret, {"ERROR": "json error"}



from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Type, Union

from bson import ObjectId
from pydantic import BaseModel, Field, validator

from fastel.exceptions import APIException


class BaseCartConfig(BaseModel):
    @classmethod
    def model_with_optional_fields(cls: Type[BaseModel]) -> Type[BaseModel]:
        """Generate a `BaseModel` class with all the same fields as `model` but as optional"""

        class OptionalModel(cls):  # type: ignore
            ...

        for field in OptionalModel.__fields__.values():
            field.required = False

        # for generated schema for example (can be removed)
        OptionalModel.__name__ = f"Optional{cls.__name__}"

        return OptionalModel

    @classmethod
    def validate_optional(cls, value: Any) -> BaseModel:
        return cls.model_with_optional_fields().validate(value)


class PaymentSubTypes(str, Enum):
    atm = "atm"
    credit = "credit"
    cod = "cod"
    offline = "offline"
    linepay = "linepay"
    barcode = "barcode"
    webatm = "webatm"
    cvs = "cvs"
    paypal = "paypal"
    unknown = "unknown"
    default = "default"  # ecpay payment all open
    apple_pay = "apple_pay"
    google_pay = "google_pay"
    express = "express"


class LogisticTypes(str, Enum):
    HOME = "HOME"
    CVS = "CVS"
    SELF_PICK = "SELF_PICK"
    CUSTOM = "CUSTOM"
    OVERSEA = "OVERSEA"


class HomeSubtypes(str, Enum):
    ECAN = "ECAN"
    TCAT = "TCAT"
    SF = "SF"


class OverseaSubtype(str, Enum):
    HK = "HK"  # 香港
    MO = "MO"  # 澳門
    CN = "CN"  # 大陸
    MY = "MY"  # 馬來西亞
    SG = "SG"  # 新加坡
    TH = "TH"  # 泰國


class InvoiceTypes(str, Enum):
    B2B = "B2B"  # 三聯
    B2C = "B2C"  # 二聯
    B2C_DONATE = "B2C_DONATE"  # 捐贈
    B2C_PHONE_CARRIER = "B2C_PHONE_CARRIER"  # 手機載具
    B2C_NPC = "B2C_NPC"  # 自然人憑證
    B2C_PROVIDER = "B2C_PROVIDER"  # 會員載具


class CVSLogisticSubTypes(str, Enum):
    FAMI = "FAMI"
    UNIMART = "UNIMART"
    HILIFE = "HILIFE"
    FAMIC2C = "FAMIC2C"
    UNIMARTC2C = "UNIMARTC2C"
    HILIFEC2C = "HILIFEC2C"
    OKMARTC2C = "OKMARTC2C"


class CartConfig(BaseCartConfig):
    buyer_name: str = ""
    buyer_phone: str = ""
    buyer_email: str = ""
    buyer_zip: str = ""
    buyer_city: str = ""
    buyer_district: str = ""
    buyer_address: str = ""
    buyer_tel: str = ""
    buyer_tel_ext: str = ""
    receiver_name: str = ""
    receiver_phone: str = ""
    receiver_email: str = ""
    receiver_zip: str = ""
    receiver_city: str = ""
    receiver_district: str = ""
    receiver_address: str = ""
    receiver_tel: str = ""
    receiver_tel_ext: str = ""
    payment_subtype: PaymentSubTypes = PaymentSubTypes.credit
    logistics_type: LogisticTypes = LogisticTypes.HOME
    logistics_subtype: str = ""
    logistics_cvs_store_id: Optional[str]

    invoice_type: InvoiceTypes = InvoiceTypes.B2C_PROVIDER
    b2b_company_name: str = ""
    b2b_company_no: str = ""
    b2c_phone_carrier_code: str = ""
    b2c_donate_code: str = ""
    b2c_npc_code: str = ""
    b2c_provider_code: str = ""

    order_note: str = ""
    points: int = 0
    gift_points: int = 0
    use_full_points: bool = False
    extra_data: Optional[Any]

    @validator("logistics_subtype")
    def validate_logistic_subtype(cls, value: str, values: Dict[str, Any]) -> str:
        if value is None:
            return ""
        try:
            if values["logistics_type"] in [
                LogisticTypes.SELF_PICK,
                LogisticTypes.CUSTOM,
            ]:
                return ""
            elif values["logistics_type"] == LogisticTypes.HOME:
                return value
            elif values["logistics_type"] == LogisticTypes.OVERSEA:
                return OverseaSubtype[value]
            return CVSLogisticSubTypes[value]
        except KeyError:
            raise ValueError("subtype_not_valid")


OptionalCartConfig = CartConfig.model_with_optional_fields()


class SingleConfig(BaseModel):
    name: str
    choice: str


class BoolConfig(BaseModel):
    name: str


class ItemConfig(BaseModel):
    qty: int
    variants: List[Union[SingleConfig, BoolConfig]] = []
    extra_data: Optional[Any]

    @validator("qty")
    def positive_qty(cls, qty: int) -> int:
        if qty < 1:
            raise APIException(
                status_code=400, error="invalid_qty_error", detail="quantity must > 0"
            )
        return qty

    @validator("variants")
    def validate_variants(
        cls, value: List[Union[SingleConfig, BoolConfig]]
    ) -> List[Union[SingleConfig, BoolConfig]]:
        names = []
        for variant in value:
            if variant.name in names:
                raise APIException(status_code=400, error="invalid_variant", detail="")
            names.append(variant.name)

        return value


class VariantTypes(str, Enum):
    bool = "bool"
    single = "single"


class SingleChoice(BaseModel):
    name: str
    label: str
    price: int


class SingleVariant(BaseModel):
    type: Literal["single"]
    name: str
    label: str
    choices: List[SingleChoice]


class BooleanVariant(BaseModel):
    type: Literal["bool"]
    name: str
    label: str
    price: int


class ValidatedObjectId(ObjectId):  # type: ignore
    @classmethod
    def __get_validators__(cls) -> Any:
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def validate(cls, value: Any) -> ObjectId:
        return ObjectId(value)


class Product(BaseModel):
    id: ValidatedObjectId = Field(alias="_id")
    name: str
    price: int
    # always 常駐商品 period 期間商品 total 庫存商品
    stock_type: Optional[Literal["always", "period", "total"]] = "always"
    stock_amount: Optional[int] = 0
    stock_sold_amount: Optional[int] = 0
    stock_start_date: Optional[str]
    stock_start_time: Optional[str]
    stock_end_date: Optional[str]
    stock_end_time: Optional[str]
    variants: List[Union[SingleVariant, BooleanVariant]] = []
    on_shelf: Optional[bool]
    public: Optional[bool]
    labels: Optional[List[str]]
    images: Optional[Union[List[str], List[Dict[str, Any]]]]
    session: Optional[Dict[str, Any]]
    extra_data: Optional[Any]
    extra_config_schema: Optional[str]
    n_in_1_products: Optional[List[str]]
    type: Optional[str]
    no_shipping_cost: Optional[bool]
    shipping_extra_price: Optional[int]
    no_global_free_shipping: Optional[bool]

    class Config:
        allow_population_by_field_name = True


class Coupon(BaseModel):
    id: ValidatedObjectId = Field(alias="_id")
    name: str
    code: str
    discount: int
    threshold: int
    start_time: int
    end_time: int
    usage: int = 0


class Discount(BaseModel):
    id: ValidatedObjectId = Field(alias="_id")
    name: str
    discount: int
    threshold: int
    start_time: int
    end_time: int



import base64
import hmac
import json
import urllib.parse
from hashlib import md5, sha256
from typing import Any, Callable, Dict, Union

from Crypto.Cipher import AES


class BaseCryptor:
    def __init__(self, hash_key: str, hash_iv: str):
        self.hash_key = hash_key
        self.hash_iv = hash_iv

    def encrypt(self, data: Any) -> str:
        raise NotImplementedError

    def decrypt(self, data: Any) -> str:
        raise NotImplementedError

    def build_trade_sha(self, built_data: str) -> str:
        raise NotImplementedError


class EcInvoiceCryptor(BaseCryptor):
    def pad(self, str_data: str) -> str:
        length = AES.block_size - (len(str_data) % AES.block_size)
        if length == 0:
            length = AES.block_size
        str_data += chr(length) * length
        return str_data

    def encrypt(self, data: Dict[str, Any]) -> str:
        json_data = json.dumps(data)
        encode_str = urllib.parse.quote(json_data)
        cryptor = AES.new(self.hash_key, AES.MODE_CBC, self.hash_iv)
        result = cryptor.encrypt(self.pad(encode_str).encode())
        return base64.b64encode(result).decode()

    def decrypt(self, data: Union[str, bytes]) -> str:
        unpad: Callable[[str], str] = lambda s: s[: -ord(s[len(s) - 1 :])]
        data = base64.b64decode(data)
        cryptor = AES.new(self.hash_key, AES.MODE_CBC, self.hash_iv)
        aes_data = cryptor.decrypt(data).decode()
        unpad_data = unpad(aes_data)
        real_data = urllib.parse.unquote(unpad_data).replace("null", '""')
        real_data = eval(real_data)
        return real_data


class SHACryptor(BaseCryptor):
    def _format_str(self, _str: str) -> str:
        conflict_patterns = [
            ("%2d", "-"),
            ("%5f", "_"),
            ("%2e", "."),
            ("%21", "!"),
            ("%2a", "*"),
            ("%28", "("),
            ("%29", ")"),
            ("%20", "+"),
            ("/", "%2f"),
        ]
        for pattern, replacement in conflict_patterns:
            _str = _str.replace(pattern, replacement)
        return _str

    def encrypt(self, data: Dict[str, Any]) -> str:
        cipher = sha256
        real_data = json.loads(json.dumps(data, sort_keys=True))
        data_str = "&".join(map(lambda v: f"{v[0]}={v[1]}", real_data.items()))
        data_str = f"HashKey={self.hash_key}&{data_str}&HashIV={self.hash_iv}".lower()
        data_str = self._format_str(urllib.parse.quote(data_str).lower())
        d = cipher(data_str.encode())
        return d.hexdigest().upper()


class MD5Cryptor(SHACryptor):
    def encrypt(self, data: Dict[str, Any]) -> str:
        cipher = md5
        real_data = json.loads(json.dumps(data, sort_keys=True))
        data_str = "&".join(map(lambda v: f"{v[0]}={v[1]}", real_data.items()))
        data_str = f"HashKey={self.hash_key}&{data_str}&HashIV={self.hash_iv}".lower()
        data_str = self._format_str(urllib.parse.quote(data_str).lower())
        d = cipher(data_str.encode())

        return d.hexdigest().upper()


class AESCryptor(BaseCryptor):
    def pad(self, str_data: str) -> str:
        length = AES.block_size - (len(str_data) % AES.block_size)
        str_data += chr(length) * length
        return str_data

    def encrypt(self, data: Dict[str, Any]) -> str:
        parse_data = {}
        for key, value in data.items():
            if value is not None:
                parse_data[key] = value
        info_str = urllib.parse.urlencode(parse_data, quote_via=urllib.parse.quote)  # type: ignore
        cryptor = AES.new(self.hash_key.encode(), AES.MODE_CBC, self.hash_iv.encode())
        result = cryptor.encrypt(self.pad(info_str).encode())
        result = result.hex()
        assert isinstance(result, str)
        return result

    def decrypt(self, data: str) -> str:
        unpad: Callable[[str], str] = lambda s: s[: -ord(s[len(s) - 1 :])]
        cyrptor = AES.new(self.hash_key.encode(), AES.MODE_CBC, self.hash_iv.encode())
        return unpad(cyrptor.decrypt(bytes.fromhex(data)).decode())

    def build_trade_sha(self, built_data: str) -> str:
        key_data_iv = "HashKey={}&{}&HashIV={}".format(
            self.hash_key, built_data, self.hash_iv
        )
        return sha256(key_data_iv.encode()).hexdigest().upper()


class HMACCryptor:
    def __init__(self, key: str):
        self.key = key

    def encrypt(self, nonce: str, uri: str, json_str: str) -> str:
        _signature = hmac.new(
            key=self.key.encode(),
            msg=(self.key + uri + json_str + nonce).encode(),
            digestmod=sha256,
        )
        signature = base64.b64encode(_signature.digest()).decode()
        return signature



def get_gift_points_amount(
    user_gift_points: int,
    gift_points_ratio: int,
    use_points_ratio: float,
    current_total: int,
    gift_points: int = 0,
) -> int:
    if gift_points > 0:
        gift_points_amount = min(
            # 計算用戶點數比例金額後的數值
            int(user_gift_points / gift_points_ratio),
            # 計算帶入點數比例金額後的數值
            int(gift_points / gift_points_ratio),
            # 與目前能夠折抵的訂單折扣後價格的金額
            round(use_points_ratio * current_total),
        )
    else:
        gift_points_amount = min(
            # 計算用戶點數比例金額後的數值
            int(user_gift_points / gift_points_ratio),
            # 與目前能夠折抵的訂單折扣後價格的金額
            round(use_points_ratio * current_total),
        )
    return gift_points_amount


def get_points_amount(
    user_points: int,
    points_ratio: int,
    use_points_ratio: float,
    current_total: int,
    points: int = 0,
) -> int:
    if points > 0:
        points_amount = min(
            # 計算用戶點數比例金額後的數值
            user_points * points_ratio,
            # 計算帶入點數比例金額後的數值
            points * points_ratio,
            # 與目前能夠折抵的訂單折扣後價格的金額
            round(use_points_ratio * current_total),
        )
    else:
        points_amount = min(
            # 計算用戶點數比例金額後的數值
            user_points * points_ratio,
            # 與目前能夠折抵的訂單折扣後價格的金額
            round(use_points_ratio * current_total),
        )
    return points_amount



import json
import os
from base64 import b64decode
from typing import Any, Callable, Dict


class SdkConfigCls:
    api_host: str = ""
    web_host: str = ""
    sdk_client: str = ""
    stage: str = ""
    client_secret: str = ""
    conversion_id: str = ""
    conversion_token: str = ""

    s3_bucket: str = ""

    sf_merchant_id: str = ""
    sf_aes_key: str = ""
    sf_app_key: str = ""
    sf_secret: str = ""
    sf_card_no: str = ""

    payment_stepfn_arn: str = ""
    logistics_stepfn_arn: str = ""

    neweb_merchant_id: str = ""
    neweb_hash_key: str = ""
    neweb_hash_iv: str = ""

    neweb_invoice_merchant_id: str = ""
    neweb_invoice_hash_key: str = ""
    neweb_invoice_hash_iv: str = ""

    ecpay_merchant_id: str = ""
    ecpay_hash_key: str = ""
    ecpay_hash_iv: str = ""

    ecpay_invoice_merchant_id: str = ""
    ecpay_invoice_hash_key: str = ""
    ecpay_invoice_hash_iv: str = ""

    ecpay_logistics_merchant_id: str = ""
    ecpay_logistics_hash_key: str = ""
    ecpay_logistics_hash_iv: str = ""

    linepay_channel_id: str = ""
    linepay_channel_secret: str = ""

    tappay_partner_key: str = ""
    tappay_merchant_id: str = ""

    public_key: Dict[str, Any]
    auth_host: str
    ntfn_host: str

    extra_config: Dict[str, Any] = {}

    @property
    def provider(self) -> str:
        parts = self.sdk_client.split(".")
        return parts[0]

    @property
    def client_id(self) -> str:
        parts = self.sdk_client.split(".")
        return parts[-1]

    @property
    def is_service(self) -> bool:
        return True if self.provider == "service" else False

    @property
    def ecpay_logistics_host(self) -> str:
        if self.stage in ["stg", "dev"]:
            return "https://logistics-stage.ecpay.com.tw"
        elif self.stage in ["prod", "production"]:
            return "https://logistics.ecpay.com.tw"
        else:
            return ""

    def put_extras(self, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            self.extra_config[key] = value

    def get_extra(self, key: str, default_value: Any = None) -> Any:
        return self.extra_config.get(key, default_value)

    def set_callable(self, key: str, callable: Callable[..., Any]) -> None:
        self.extra_config[key] = callable

    def get_callable(
        self, key: str, default_callable: Callable[..., Any]
    ) -> Callable[..., Any]:
        return self.extra_config.get(key, default_callable)


SdkConfig = SdkConfigCls()


def sdk_auto_config(sdk_client: str = "") -> None:
    global SdkConfig
    print("SdkConfig", SdkConfig)

    def get_env_or_raise(key: str, raise_exception: bool = True) -> str:
        value: str = os.environ.get(key, "")
        if value == "" and raise_exception:
            raise AttributeError(f"[sdk_auto_config] {key} not found")
        return value

    # if not sdk_client initial app by revtel
    if not sdk_client:
        client_id = get_env_or_raise("CLIENT_ID")
        sdk_client = f"app.{client_id}"

    # common env vars
    SdkConfig.stage = get_env_or_raise("STAGE")

    b64_public = get_env_or_raise("PUBLIC_KEY")
    SdkConfig.public_key = json.loads(b64decode(b64_public.encode("UTF-8")).decode())
    SdkConfig.auth_host = get_env_or_raise("AUTH_HOST", False)
    SdkConfig.ntfn_host = get_env_or_raise("NTFN_HOST", False)

    SdkConfig.sdk_client = sdk_client

    if not SdkConfig.is_service:
        SdkConfig.client_secret = get_env_or_raise("CLIENT_SECRET")
        # for facebook conversion vars
        SdkConfig.conversion_id = get_env_or_raise(
            "CONVERSION_ID", raise_exception=False
        )
        SdkConfig.conversion_token = get_env_or_raise(
            "CONVERSION_TOKEN", raise_exception=False
        )
        SdkConfig.payment_stepfn_arn = get_env_or_raise("PAYMENT_STEPFN_ARN", False)
        SdkConfig.logistics_stepfn_arn = get_env_or_raise("LOGISTICS_STEPFN_ARN", False)
        SdkConfig.neweb_merchant_id = get_env_or_raise("NEWEB_MERCHANT_ID", False)
        SdkConfig.neweb_hash_key = get_env_or_raise("NEWEB_HASH_KEY", False)
        SdkConfig.neweb_hash_iv = get_env_or_raise("NEWEB_HASH_IV", False)

        SdkConfig.neweb_invoice_merchant_id = get_env_or_raise(
            "NEWEB_INVOICE_MERCHANT_ID", False
        )
        SdkConfig.neweb_invoice_hash_key = get_env_or_raise(
            "NEWEB_INVOICE_HASH_KEY", False
        )
        SdkConfig.neweb_invoice_hash_iv = get_env_or_raise(
            "NEWEB_INVOICE_HASH_IV", False
        )

        SdkConfig.ecpay_merchant_id = get_env_or_raise("ECPAY_MERCHANT_ID", False)
        SdkConfig.ecpay_hash_key = get_env_or_raise("ECPAY_HASH_KEY", False)
        SdkConfig.ecpay_hash_iv = get_env_or_raise("ECPAY_HASH_IV", False)

        SdkConfig.ecpay_invoice_merchant_id = get_env_or_raise(
            "ECPAY_INVOICE_MERCHANT_ID", False
        )
        SdkConfig.ecpay_invoice_hash_key = get_env_or_raise(
            "ECPAY_INVOICE_HASH_KEY", False
        )
        SdkConfig.ecpay_invoice_hash_iv = get_env_or_raise(
            "ECPAY_INVOICE_HASH_IV", False
        )
        SdkConfig.linepay_channel_id = get_env_or_raise("LINEPAY_CHANNEL_ID", False)
        SdkConfig.linepay_channel_secret = get_env_or_raise(
            "LINEPAY_CHANNEL_SECRET", False
        )

        SdkConfig.ecpay_logistics_merchant_id = get_env_or_raise(
            "ECPAY_LOGISTICS_MERCHANT_ID", False
        )
        SdkConfig.ecpay_logistics_hash_key = get_env_or_raise(
            "ECPAY_LOGISTICS_HASH_KEY", False
        )
        SdkConfig.ecpay_logistics_hash_iv = get_env_or_raise(
            "ECPAY_LOGISTICS_HASH_IV", False
        )

        SdkConfig.api_host = get_env_or_raise("API_HOST", False)
        SdkConfig.web_host = get_env_or_raise("WEB_HOST", False)
        SdkConfig.s3_bucket = get_env_or_raise("S3_BUCKET", False)
        SdkConfig.sf_merchant_id = get_env_or_raise("SF_MERCHANT_ID", False)
        SdkConfig.sf_aes_key = get_env_or_raise("SF_AES_KEY", False)
        SdkConfig.sf_app_key = get_env_or_raise("SF_APP_KEY", False)
        SdkConfig.sf_secret = get_env_or_raise("SF_SECRET", False)
        SdkConfig.sf_card_no = get_env_or_raise("SF_CARD_NO", False)

        SdkConfig.tappay_merchant_id = get_env_or_raise("TAPPAY_MERCHANT_ID", False)
        SdkConfig.tappay_partner_key = get_env_or_raise("TAPPAY_PARTNER_KEY", False)



from typing import Any, List, Literal, Optional

from pydantic import BaseModel


class CheckoutProduct(BaseModel):
    id: str
    name: str
    imageUrl: Optional[str]
    quantity: int
    price: int


class CheckoutPackage(BaseModel):
    id: str
    amount: int
    products: List[CheckoutProduct]


class CheckoutRedirectUrl(BaseModel):
    confirmUrl: str
    cancelUrl: str


class LinePayCheckoutModel(BaseModel):
    amount: int
    currency: Literal["USD", "JPY", "TWD", "THB"] = "TWD"
    orderId: str
    packages: List[CheckoutPackage]
    redirectUrls: CheckoutRedirectUrl


class LinePayCallbackModel(BaseModel):
    orderId: str
    transactionId: str


class LinePayConfirmModel(BaseModel):
    amount: int
    currency: Literal["USD", "JPY", "TWD", "THB"] = "TWD"
    transactionId: str


class LinePayTransactionDetail(BaseModel):
    transaction_id: str
    order_id: str


class LinePayResponse(BaseModel):
    returnCode: str
    returnMessage: str
    info: Any


class LinePayRequestPaymentUrl(BaseModel):
    app: str
    web: str


class LinePayRequestInfo(BaseModel):
    transactionId: str
    paymentAccessToken: str
    paymentUrl: LinePayRequestPaymentUrl


class LinePayRequestResp(LinePayResponse):
    info: Optional[LinePayRequestInfo]


class LinePayConfirmInfo(BaseModel):
    orderId: str
    transactionId: str
    authorizationExpireDate: Optional[str]
    regKey: Optional[str]
    payInfo: Optional[List[Any]]
    packages: Optional[List[Any]]
    merchantReference: Optional[Any]
    shipping: Optional[Any]


class LinePayConfirmResp(LinePayResponse):
    info: Optional[LinePayConfirmInfo]



from datetime import datetime, timezone
from enum import Enum


class PaymentStatus(str, Enum):
    PENDING = "pending"
    WAITING = "waiting"
    SUCCESS = "success"
    FAILURE = "failure"
    CODE_GENERATED = "code_generated"
    REFUNDED = "refunded"


def auto_order_number() -> str:
    order_number = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S%f")[:-4]
    return order_number


def parse_order_number(order_number: str) -> str:
    if "_" in order_number:
        return order_number.split("_")[0]
    return order_number[:16]



import io

import boto3

from fastel.config import SdkConfig
from fastel.logistics.common.helpers import get_shipping_note_method
from fastel.logistics.common.models import LogisticsDataModel


def upload_shipping_note(logistics_data: LogisticsDataModel) -> None:
    try:
        method = get_shipping_note_method("ecpay")
        shipping_note, content_type = method(logistics_data.dict(exclude_none=True))
        s3 = boto3.client("s3", "ap-northeast-1")
        buf = io.BytesIO()
        buf.write(shipping_note)
        buf.seek(0)
        s3.upload_fileobj(
            buf,
            SdkConfig.s3_bucket,
            f"shipping_note/{logistics_data.logistics_id}",
            # shipping note must be private file
            ExtraArgs={"ContentType": content_type},
        )
        print(
            "[UPLOAD SHIPPING] ------------------------- SUCCESS ------------------------"
        )
    except Exception as error:
        print(
            "[UPLOAD SHIPPING] ------------------------- FAIL ------------------------"
        )
        print(error)
        raise



import hashlib
import hmac
from typing import Any, Dict, Set

from fastel.utils import requests


class TrackingClient:
    url = "https://api.trackingmore.com/v2"

    def __init__(self, api_key: str) -> None:
        self._api_key = api_key

    def _get(self, path: str) -> Dict[str, Any]:
        headers = {
            "Content-Type": "application/json",
            "Trackingmore-Api-Key": self._api_key,
        }
        full_url = f"{self.url}{path}"
        resp = requests.get(url=full_url, headers=headers)
        resp.raise_for_status()
        return resp.json()  # type: ignore

    def _post(self, path: str, data: Dict[str, Any]) -> Dict[str, Any]:
        headers = {
            "Content-Type": "application/json",
            "Trackingmore-Api-Key": self._api_key,
        }
        full_url = f"{self.url}{path}"
        resp = requests.post(url=full_url, json=data, headers=headers)
        resp.raise_for_status()
        return resp.json()  # type: ignore

    def create_tracking(self, logistics_id: str, carrier_code: str) -> Dict[str, Any]:
        data = {"tracking_number": logistics_id, "carrier_code": carrier_code}
        resp = self._post("/trackings/post", data=data)
        return resp

    def get_tracking(self, logistics_id: str, carrier_code: str) -> Dict[str, Any]:
        path = f"/trackings/{carrier_code}/{logistics_id}"
        resp = self._get(path)
        return resp

    def verify_signature(self, timestamp: int, signature: str, key: str) -> None:
        bytes_timestamp = str(timestamp).encode()
        bytes_key = key.encode()
        sign = hmac.new(
            key=bytes_key, msg=bytes_timestamp, digestmod=hashlib.sha256
        ).hexdigest()
        if sign != signature:
            raise ValueError("invalid signature")

    def verify(self, timestamp: int, signature: str, keys: Set[str]) -> bool:
        for key in keys:
            try:
                self.verify_signature(timestamp, signature, key)
                return True
            except ValueError:
                continue
        raise ValueError("invalid signature")



from typing import Any, Dict, Optional
from uuid import uuid4

import boto3

from fastel.cart import Cart
from fastel.cart.datastructures import PaymentSubTypes
from fastel.collections import get_collection
from fastel.config import SdkConfig
from fastel.exceptions import APIException
from fastel.payment.common.models.callback import CallbackModel
from fastel.payment.tappay.gateway import TappayGateway
from fastel.payment.utils import PaymentStatus, auto_order_number
from fastel.utils import timestamp_postfix


class TappayHelper:
    def __init__(self, partner_key: str, merchant_id: str, stage: str) -> None:
        self.gateway = TappayGateway(
            partner_key=partner_key, merchant_id=merchant_id, stage=stage
        )

    def handle_callback(self, resp: Dict[str, Any]) -> None:
        status = resp.get("status", None)

        checkout_col = get_collection("checkout")
        checkout = checkout_col.find_one({"order_number": resp["order_number"]})

        if not checkout:
            raise APIException(status_code=400, error="invalid_order", detail="")

        if status == 0:
            payment_status = PaymentStatus.SUCCESS
        else:
            payment_status = PaymentStatus.FAILURE

        if payment_status == PaymentStatus.SUCCESS and "card_secret" in resp:
            card_secret = resp["card_secret"]
            card_last_four = resp["card_info"]["last_four"]

            profile_col = get_collection("private_profile")

            profile_col.update_one(
                {"owner": checkout["owner"]},
                {
                    "$set": {
                        f'credit_cards.{card_secret["card_key"]}': {
                            **card_secret,
                            "card_last_four": card_last_four,
                        }
                    }
                },
            )

        stepfn = boto3.client("stepfunctions")
        callback_data = CallbackModel(
            order_id=checkout["order_number"],
            total=checkout["total"],
            payment_subtype=checkout["payment_subtype"],
            payment_status=payment_status,
            payment_transaction_detail=resp,
            payment_transaction_id=resp["rec_trade_id"],
        )
        stepfn.start_execution(
            stateMachineArn=SdkConfig.payment_stepfn_arn,
            name=timestamp_postfix(
                f"callback_{callback_data.order_id}_{callback_data.payment_status}"
            ),
            input=callback_data.json(),
        )

    def handle_3d_resp(self, resp: Dict[str, Any]) -> Dict[str, Any]:
        status = resp.get("status", None)

        if status == 0:
            return {"url": resp["payment_url"]}

        raise APIException(
            status_code=400, error="checkout_error", detail="invalid status"
        )

    def handle_n3d_resp(
        self, checkout_dict: Dict[str, Any], resp: Dict[str, Any]
    ) -> Dict[str, Any]:
        status = resp.get("status", None)

        if status == 0:
            payment_status = PaymentStatus.SUCCESS
        else:
            payment_status = PaymentStatus.FAILURE

        callback_data = CallbackModel(
            order_id=checkout_dict["order_number"],
            total=checkout_dict["total"],
            payment_subtype=checkout_dict["payment_subtype"],
            payment_status=payment_status,
            payment_transaction_detail=resp,
            payment_transaction_id=resp["rec_trade_id"],
        )

        stepfn = boto3.client("stepfunctions")
        stepfn.start_execution(
            stateMachineArn=SdkConfig.payment_stepfn_arn,
            name=timestamp_postfix(
                f"callback_{callback_data.order_id}_{callback_data.payment_status}"
            ),
            input=callback_data.json(),
        )

        checkout_dict["payment_status"] = payment_status

        return json.loads(checkout_dict)  # type: ignore

    def refund(self, order: Dict[str, Any]) -> Dict[str, Any]:
        if order["payment_status"] != PaymentStatus.SUCCESS:
            raise APIException(status_code=400, error="invalid_order", detail="")

        refund_resp = self.gateway.refund(order["payment_info"]["rec_trade_id"])
        if refund_resp["status"] != 0:
            raise APIException(status_code=400, error="refund_failed", detail="")

        order = get_collection("order").find_one_and_update(
            {"_id": order["_id"]},
            {
                "$set": {
                    "payment_status": PaymentStatus.REFUNDED,
                    "refund_detail": refund_resp,
                }
            },
            return_document=True,
        )

        return order

    def token_request(
        self,
        cart: Cart,
        card_token: str,
        card_key: str,
    ) -> Dict[str, Any]:
        total = cart.total
        item_names = [item.product.name for item in cart.items]
        details = " | ".join(item_names)
        order_number = auto_order_number()

        checkout = cart.to_checkout(order_number=order_number)

        result_url = None
        resp = self.gateway.token_request(
            amount=total,
            order_number=order_number,
            details=details,
            card_key=card_key,
            card_token=card_token,
            card_ccv=None,
            three_domain_secure=False,
            result_url=result_url,
        )
        return self.handle_n3d_resp(checkout, resp)

    def prime_request(
        self,
        cart: Cart,
        prime: str,
        redirect_url: Optional[str] = None,
        callback_url: Optional[str] = None,
        cancel_url: Optional[str] = None,
    ) -> Dict[str, Any]:
        total = cart.total
        item_names = [item.product.name for item in cart.items]
        details = " | ".join(item_names)
        order_number = auto_order_number()

        checkout = cart.to_checkout(order_number=order_number)

        three_domain_secure = False
        result_url = None
        state_ins = get_collection("state")
        r = state_ins.insert_one({"state": str(uuid4())})

        if redirect_url is None:
            redirect_url = f"{SdkConfig.api_host}/payment/tappay/redirect"

        if callback_url is None:
            callback_url = f"{SdkConfig.api_host}/payment/tappay/callback?state={str(r.inserted_id)}"

        if cancel_url is None:
            cancel_url = SdkConfig.web_host

        if cart.payment_subtype == PaymentSubTypes.credit:
            three_domain_secure = True
            result_url = {
                "frontend_redirect_url": redirect_url,
                "backend_notify_url": callback_url,
                "go_back_url": cancel_url,
            }

        cardholder = {
            "phone_number": cart.buyer_phone,
            "name": cart.buyer_name,
            "email": cart.buyer_email,
        }

        resp = self.gateway.prime_request(
            amount=total,
            order_number=order_number,
            details=details,
            prime=prime,
            remember=True,
            three_domain_secure=three_domain_secure,
            result_url=result_url,
            cardholder=cardholder,
        )

        if cart.payment_subtype == PaymentSubTypes.credit:
            return self.handle_3d_resp(resp)

        else:
            return self.handle_n3d_resp(checkout, resp)



from datetime import datetime
from typing import Any, Dict, List, Optional, Set, Tuple

from .base import BaseCart
from .collections import get_collection
from .datastructures import CartConfig
from .product_items import ProductItems
from .utils import get_gift_points_amount, get_points_amount


class Cart(BaseCart):
    product_multi_item_cls = ProductItems
    config_cls = CartConfig

    def _init_cart_extra(self) -> Dict[str, Any]:
        return {}

    def _calc_fee(self) -> Tuple[int, List[Dict[str, Any]]]:
        fee_items: List[Dict[str, Any]] = [
            {
                "name": "運費",
                "amount": 0,
            }
        ]
        return (sum(i["amount"] for i in fee_items), fee_items)

    def _calc_coupon_discounts(
        self, subtotal: int, coupon_code: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        items: List[Any] = []
        now = datetime.now().timestamp() * 1000

        if coupon_code is None:
            return items

        coupon = get_collection("coupon").find_one(
            {
                "code": coupon_code,
                "start_time": {"$lte": now},
                "end_time": {"$gte": now},
                "threshold": {"$lte": subtotal},
                "is_void": {"$ne": True},
            }
        )

        if coupon:
            used_users = coupon.pop("users", [])
            owner = self._cache_cart["owner"]
            coupon_items = {
                "type": "coupon",
                "name": coupon["name"],
                "sales_amount": round(coupon["discount"] / self._tax_ratio, 7),
                "amount": coupon["discount"],
                "coupon": coupon,
            }

            if owner in used_users:
                return items

            if coupon.get("type", "personal") == "personal":
                assigned_user_id = coupon.get("user_id", None)
                if assigned_user_id == owner:
                    items.append(coupon_items)
            else:
                if len(used_users) <= coupon["usage"]:
                    items.append(coupon_items)

        return items

    def _calc_threshold_discounts(self, subtotal: int) -> List[Dict[str, Any]]:
        items = []
        now = datetime.now().timestamp() * 1000

        if get_collection("discount") is not None:
            discounts = get_collection("discount").find(
                {
                    "start_time": {"$lte": now},
                    "end_time": {"$gte": now},
                    "threshold": {"$lte": subtotal},
                }
            )

            if discounts:
                discounts.sort((("threshold", -1),))
                for discount in discounts:
                    items.append(
                        {
                            "type": "discount",
                            "name": discount["name"],
                            "sales_amount": round(
                                discount["discount"] / self._tax_ratio, 7
                            ),
                            "amount": discount["discount"],
                            "discount": discount,
                        }
                    )
                    break

        return items

    def _calc_extra_discount(self, subtotal: int) -> List[Dict[str, Any]]:
        items: List[Dict[str, Any]] = []
        return items

    def _calc_points_discount(
        self, subtotal: int, discount_amount: int
    ) -> List[Dict[str, Any]]:
        if not self._user:
            return []
        items = []
        current_total = subtotal - discount_amount
        if current_total < 0:
            current_total = 0
        private_profile = get_collection("private_profile").find_one(
            {"owner": str(self._user["_id"])}
        )
        points_amount = 0
        gift_points_amount = 0

        user_points = private_profile.get("points", 0)
        user_gift_points = private_profile.get("gift_points", 0)
        if self.use_full_points:
            if user_gift_points > 0:
                # 使用 _gift_points_ratio 1金額：n點數 比例計算
                gift_points_amount = get_gift_points_amount(
                    user_gift_points=user_gift_points,
                    gift_points_ratio=self._gift_points_ratio,
                    use_points_ratio=self._use_points_ratio,
                    current_total=current_total,
                )
                current_total -= gift_points_amount
            if user_points > 0:
                # 使用 _points_ratio 1點數：n金額 比例計算
                points_amount = get_points_amount(
                    user_points=user_points,
                    points_ratio=self._points_ratio,
                    use_points_ratio=self._use_points_ratio,
                    current_total=current_total,
                )
        else:
            if self.gift_points > 0:
                # 使用 _gift_points_ratio 1金額：n點數 比例計算
                gift_points_amount = get_gift_points_amount(
                    user_gift_points=user_gift_points,
                    gift_points_ratio=self._gift_points_ratio,
                    use_points_ratio=self._use_points_ratio,
                    current_total=current_total,
                    gift_points=self.gift_points,
                )
                current_total -= gift_points_amount
            if self.points > 0:
                # 使用 _points_ratio 1點數：n金額 比例計算
                points_amount = get_points_amount(
                    user_points=user_points,
                    points_ratio=self._points_ratio,
                    use_points_ratio=self._use_points_ratio,
                    current_total=current_total,
                    points=self.points,
                )

        gift_points = gift_points_amount * self._gift_points_ratio

        points = (
            points_amount % self._points_ratio
            and int(points_amount / self._points_ratio) + 1
            or int(points_amount / self._points_ratio)
        )
        if gift_points > 0:
            items.append(
                {
                    "type": "gift_points",
                    "name": "紅利點數折扣",
                    "sales_amount": round(gift_points_amount / self._tax_ratio, 7),
                    "amount": gift_points_amount,
                    "points": gift_points,
                }
            )
        if points > 0:
            items.append(
                {
                    "type": "points",
                    "name": "點數折扣",
                    "sales_amount": round(points_amount / self._tax_ratio, 7),
                    "amount": points_amount,
                    "points": points,
                }
            )

        return items

    def _calc_combine_discounts(self) -> List[Dict[str, Any]]:
        """
        1. splited all items with qty
        2. traverse all splited item and compare with others
        3. get all product label combination
        4. will pass if product id is same (will not compare itself)
        """

        site_config = get_collection("site").find_one({"name": "config"}) or {}
        combine_discounts = site_config.get("combine_discount", [])
        if not combine_discounts:
            return []

        combine_discounts.sort(key=lambda c: c["discount"], reverse=True)

        discount_items = []

        splited_items = []
        for item in self.items:
            splited_items += [item] * item.config.qty

        def _get_product_label_combination(p1: Any, p2: Any) -> List[Set[Any]]:
            combinations = []
            for l1 in p1.product.labels:
                for l2 in p2.product.labels:
                    if {l1, l2} not in combinations:
                        combinations.append({l1, l2})
            return combinations

        mapped: Dict[str, Any] = {}
        while splited_items:
            cur_item = splited_items.pop()
            for index, item in enumerate(splited_items):
                b = False
                if item.product.id == cur_item.product.id:
                    continue

                for dis in combine_discounts:
                    dis_labels = set(dis["labels"])
                    combinations = _get_product_label_combination(item, cur_item)
                    if dis_labels in combinations:
                        if dis["name"] in mapped:
                            mapped[dis["name"]] += dis["discount"]
                        else:
                            mapped[dis["name"]] = dis["discount"]

                        splited_items.pop(index)
                        b = True
                        break

                if b:
                    break

        if mapped:
            for name, amount in mapped.items():
                discount_items.append(
                    {
                        "type": "combine_discount",
                        "name": name,
                        "sales_amount": round(amount / self._tax_ratio, 7),
                        "amount": amount,
                    }
                )

        return discount_items

    def _calc_discounts(
        self,
        subtotal: int,
        coupon_code: Optional[str] = None,
    ) -> Tuple[int, List[Dict[str, Any]]]:
        items = []

        items += self._calc_combine_discounts()
        items += self._calc_coupon_discounts(subtotal, coupon_code)
        items += self._calc_threshold_discounts(subtotal)
        items += self._calc_extra_discount(subtotal)
        items += self._calc_points_discount(
            subtotal, sum([item["amount"] for item in items])
        )
        return sum([item["amount"] for item in items]), items

    def _calc_addon(self, **kwargs: Any) -> Any:
        return 0, [], 0, []

    def to_checkout(self, **kwargs: Any) -> Any:
        checkout_dict = {**kwargs, **self.get_cart_ins()}
        checkout_dict.pop("_id", None)
        inserted = get_collection("checkout").insert_one(checkout_dict)
        checkout_dict["_id"] = inserted.inserted_id
        return checkout_dict



from typing import Any, Dict, Optional

import requests


class TappayGateway:
    @property
    def url(self) -> str:
        if self.stage in ["prod", "PROD"]:
            return "https://prod.tappaysdk.com/tpc"
        return "https://sandbox.tappaysdk.com/tpc"

    def __init__(self, partner_key: str, merchant_id: str, stage: str):
        self.partner_key = partner_key
        self.merchant_id = merchant_id
        self.stage = stage

    def build_headers(self) -> Dict[str, Any]:
        return {"Content-Type": "application/json", "x-api-key": self.partner_key}

    def prime_request(
        self,
        amount: int,
        order_number: str,
        details: str,
        prime: str,
        cardholder: Dict[str, Any],
        three_domain_secure: bool,
        result_url: Optional[Dict[str, Any]] = None,
        currency: str = "TWD",
        remember: bool = True,
    ) -> Dict[str, Any]:
        endpoint = f"{self.url}/payment/pay-by-prime"
        body = {
            "partner_key": self.partner_key,
            "merchant_id": self.merchant_id,
            "amount": amount,
            "order_number": order_number,
            "details": details,
            "prime": prime,
            "cardholder": cardholder,
            "three_domain_secure": three_domain_secure,
            "result_url": result_url,
            "currency": currency,
            "remember": remember,
        }
        headers = self.build_headers()
        res = requests.post(url=endpoint, json=body, headers=headers)
        res.raise_for_status()
        res_json = res.json()
        assert isinstance(res_json, dict)
        return res_json

    def token_request(
        self,
        amount: int,
        order_number: str,
        details: str,
        card_key: str,
        card_token: str,
        card_ccv: Optional[str],
        three_domain_secure: bool,
        result_url: Optional[Dict[str, Any]] = None,
        currency: str = "TWD",
        remember: bool = True,
    ) -> Dict[str, Any]:
        endpoint = f"{self.url}/payment/pay-by-token"
        body = {
            "partner_key": self.partner_key,
            "merchant_id": self.merchant_id,
            "amount": amount,
            "order_number": order_number,
            "details": details,
            "card_key": card_key,
            "card_token": card_token,
            "card_ccv": card_ccv,
            "three_domain_secure": three_domain_secure,
            "result_url": result_url,
            "currency": currency,
            "remember": remember,
        }
        headers = self.build_headers()
        res = requests.post(url=endpoint, json=body, headers=headers)
        res.raise_for_status()
        res_json = res.json()
        assert isinstance(res_json, dict)
        return res_json

    def get_record(self, rec_trade_id: str, order_number: str) -> Dict[str, Any]:
        endpoint = f"{self.url}/transaction/query"
        request_body = {
            "partner_key": self.partner_key,
            "records_per_page": 10,
            "page": 0,
            "filters": {"rec_trade_id": rec_trade_id, "order_number": order_number},
        }
        headers = self.build_headers()
        res = requests.post(url=endpoint, json=request_body, headers=headers)
        res.raise_for_status()
        res_json = res.json()
        assert isinstance(res_json, dict)
        return res_json

    def refund(self, rec_trade_id: str) -> Dict[str, Any]:
        endpoint = f"{self.url}/transaction/refund"
        body = {
            "partner_key": self.partner_key,
            "rec_trade_id": rec_trade_id,
        }
        resp = requests.post(url=endpoint, json=body, headers=self.build_headers())
        resp.raise_for_status()
        resp_json = resp.json()
        return resp_json  # type: ignore
